# This file is part of tf-mdp.

# tf-mdp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-mdp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-mdp. If not, see <http://www.gnu.org/licenses/>.

import rddlgym
from rddl2tf.compiler import Compiler

from tfmdp.train.policy import DeepReactivePolicy
from tfmdp.train.mrm import MarkovCell, MarkovRecurrentModel, ReparameterizationType

import numpy as np
import tensorflow as tf

import unittest


class TestMarkovCell(unittest.TestCase):

    def setUp(self):
        self.rddl1 = rddlgym.make('Navigation-v3', mode=rddlgym.AST)
        self.compiler1 = Compiler(self.rddl1, batch_mode=True)

        self.layers = [64, 32, 16]
        self.policy1 = DeepReactivePolicy(self.compiler1, self.layers, tf.nn.elu, input_layer_norm=False)

        self.batch_size1 = 8
        self.cell1 = MarkovCell(self.compiler1, self.policy1, self.batch_size1)

    def test_state_size(self):
        expected1 = ((2,),)
        actual1 = self.cell1.state_size
        self.assertTupleEqual(actual1, expected1)

    def test_output_size(self):
        expected1 = (((2,),), ((2,),), ((2,), (2,)), 1, 1)
        actual1 = self.cell1.output_size
        self.assertTupleEqual(actual1, expected1)

    def test_initial_state(self):
        cells = [self.cell1]
        batch_sizes = [self.batch_size1]
        for cell, batch_size in zip(cells, batch_sizes):
            initial_state = cell.initial_state()
            self.assertIsInstance(initial_state, tuple)
            self.assertEqual(len(initial_state), len(cell.state_size))
            for t, shape in zip(initial_state, cell.state_size):
                self.assertIsInstance(t, tf.Tensor)
                expected_shape = [batch_size] + list(shape)
                if len(expected_shape) == 1:
                    expected_shape += [1]
                self.assertListEqual(t.shape.as_list(), expected_shape)

    def test_output_simulation_step(self):
        horizon = 40
        cells = [self.cell1]
        batch_sizes = [self.batch_size1]
        for cell, batch_size in zip(cells, batch_sizes):
            with cell.graph.as_default():
                # initial_state
                initial_state = cell.initial_state()

                # timestep
                timestep = tf.constant(horizon, dtype=tf.float32)
                timestep = tf.expand_dims(timestep, -1)
                timestep = tf.stack([timestep] * batch_size)

                # reparam_flag
                reparam_flag = tf.constant(0.0, shape=(batch_size,1), dtype=tf.float32)
                input = tf.concat([timestep, reparam_flag], axis=1)

                # simulation step
                output, _ = cell(input, initial_state)
                self.assertIsInstance(output, tuple)
                self.assertEqual(len(output), 5)

    def test_next_state_in_not_reparameterized_cell(self):
        grad_a, grad_s = self._test_grad_next_state(MarkovRecurrentModel.NOT_REPARAMETERIZED_FLAG)
        self.assertTrue(np.sum(grad_a) == 0.0)
        self.assertTrue(np.sum(grad_s) == 0.0)

    def test_next_state_in_fully_reparameterized_cell(self):
        grad_a, grad_s = self._test_grad_next_state(MarkovRecurrentModel.FULLY_REPARAMETERIZED_FLAG)
        self.assertTrue(np.sum(grad_a) != 0.0)
        self.assertTrue(np.sum(grad_s) != 0.0)

    def _test_grad_next_state(self, reparameterization_type):
        horizon = 40
        cells = [self.cell1]
        batch_sizes = [self.batch_size1]
        for cell, batch_size in zip(cells, batch_sizes):
            with cell.graph.as_default():
                # initial_state
                initial_state = cell.initial_state()

                # timestep
                timestep = tf.constant(horizon, dtype=tf.float32)
                timestep = tf.expand_dims(timestep, -1)
                timestep = tf.stack([timestep] * batch_size)

                # reparam_flag
                reparam_flag = tf.constant(reparameterization_type, shape=(batch_size,1), dtype=tf.float32)
                input = tf.concat([timestep, reparam_flag], axis=1)

                # simulation step
                output, _ = cell(input, initial_state)
                next_state, action, _, _, _ = output

                loss = tf.reduce_sum(next_state[0])
                grad_a, grad_s = tf.gradients(ys=loss, xs=[action[0], initial_state[0]])

                with tf.Session() as sess:
                    sess.run(tf.global_variables_initializer())
                    return sess.run([grad_a, grad_s])

    def test_next_state_output(self):
        horizon = 40
        cells = [self.cell1]
        batch_sizes = [self.batch_size1]
        for cell, batch_size in zip(cells, batch_sizes):
            with cell.graph.as_default():
                # initial_state
                initial_state = cell.initial_state()

                # timestep
                timestep = tf.constant(horizon, dtype=tf.float32)
                timestep = tf.expand_dims(timestep, -1)
                timestep = tf.stack([timestep] * batch_size)

                # reparam_flag
                reparam_flag = tf.constant(0.0, shape=(batch_size,1), dtype=tf.float32)

                input = tf.concat([timestep, reparam_flag], axis=1)

                # simulation step
                output, _ = cell(input, initial_state)
                next_state, _, _, _, _ = output
                next_state_size, _, _, _, _ = cell.output_size

                self.assertIsInstance(next_state, tuple)
                self.assertEqual(len(next_state), len(next_state_size))
                for s, sz in zip(next_state, next_state_size):
                    self.assertIsInstance(s, tf.Tensor)
                    self.assertEqual(s.dtype, tf.float32)
                    self.assertListEqual(s.shape.as_list(), [batch_size] + list(sz))

    def test_action_output(self):
        horizon = 40
        cells = [self.cell1]
        batch_sizes = [self.batch_size1]
        for cell, batch_size in zip(cells, batch_sizes):
            with cell.graph.as_default():
                # initial_state
                initial_state = cell.initial_state()

                # timestep
                timestep = tf.constant(horizon, dtype=tf.float32)
                timestep = tf.expand_dims(timestep, -1)
                timestep = tf.stack([timestep] * batch_size)

                # reparam_flag
                reparam_flag = tf.constant(0.0, shape=(batch_size,1), dtype=tf.float32)

                input = tf.concat([timestep, reparam_flag], axis=1)

                # simulation step
                output, _ = cell(input, initial_state)
                _, action, _, _, _ = output
                _, action_size, _, _, _ = cell.output_size

                self.assertIsInstance(action, tuple)
                self.assertEqual(len(action), len(action_size))
                for a, sz in zip(action, action_size):
                    self.assertIsInstance(a, tf.Tensor)
                    self.assertEqual(a.dtype, tf.float32)
                    self.assertListEqual(a.shape.as_list(), [batch_size] + list(sz))

    def test_reward_output(self):
        horizon = 40
        cells = [self.cell1]
        batch_sizes = [self.batch_size1]
        for cell, batch_size in zip(cells, batch_sizes):
            with cell.graph.as_default():
                # initial_state
                initial_state = cell.initial_state()

                # timestep
                timestep = tf.constant(horizon, dtype=tf.float32)
                timestep = tf.expand_dims(timestep, -1)
                timestep = tf.stack([timestep] * batch_size)

                # reparam_flag
                reparam_flag = tf.constant(0.0, shape=(batch_size,1), dtype=tf.float32)

                input = tf.concat([timestep, reparam_flag], axis=1)

                # simulation step
                output, _ = cell(input, initial_state)
                _, _, _, reward, _ = output
                _, _, _, reward_size, _ = cell.output_size

                self.assertIsInstance(reward, tf.Tensor)
                self.assertListEqual(reward.shape.as_list(), [batch_size, reward_size])

    def test_log_prob_not_reparameterized_cell(self):
        _test_log_prob_output(MarkovRecurrentModel.NOT_REPARAMETERIZED_FLAG)

    def test_log_prob_fully_reparameterized_cell(self):
        _test_log_prob_output(MarkovRecurrentModel.FULLY_REPARAMETERIZED_FLAG)

    def _test_log_prob_output(self, reparameterization_type):
        horizon = 40
        cells = [self.cell1]
        batch_sizes = [self.batch_size1]
        for cell, batch_size in zip(cells, batch_sizes):
            with cell.graph.as_default():
                # initial_state
                initial_state = cell.initial_state()

                # timestep
                timestep = tf.constant(horizon, dtype=tf.float32)
                timestep = tf.expand_dims(timestep, -1)
                timestep = tf.stack([timestep] * batch_size)

                # reparam_flag
                reparam_flag = tf.constant(reparameterization_type, shape=(batch_size,1), dtype=tf.float32)

                input = tf.concat([timestep, reparam_flag], axis=1)

                # simulation step
                output, _ = cell(input, initial_state)
                _, _, _, _, log_prob = output
                _, _, _, _, log_prob_size = cell.output_size

                self.assertIsInstance(log_prob, tf.Tensor)
                self.assertListEqual(log_prob.shape.as_list(), [batch_size, log_prob_size])

    def test_log_prob_fully_reparameterized_cell(self):
        self._test_grad_log_prob(MarkovRecurrentModel.FULLY_REPARAMETERIZED_FLAG)

    def test_log_prob_not_reparameterized_cell(self):
        self._test_grad_log_prob(MarkovRecurrentModel.NOT_REPARAMETERIZED_FLAG)

    def _test_grad_log_prob(self, reparameterization_type):
        horizon = 40
        cells = [self.cell1]
        batch_sizes = [self.batch_size1]
        for cell, batch_size in zip(cells, batch_sizes):
            with cell.graph.as_default():
                # initial_state
                initial_state = cell.initial_state()

                # timestep
                timestep = tf.constant(horizon, dtype=tf.float32)
                timestep = tf.expand_dims(timestep, -1)
                timestep = tf.stack([timestep] * batch_size)

                # reparam_flag
                reparam_flag = tf.constant(reparameterization_type, shape=(batch_size,1), dtype=tf.float32)
                input = tf.concat([timestep, reparam_flag], axis=1)

                # simulation step
                output, _ = cell(input, initial_state)
                next_state, action, _, _, log_prob = output

                loss = tf.reduce_sum(log_prob)
                grad_a, grad_s, grad_ss = tf.gradients(ys=loss, xs=[action[0], initial_state[0], next_state[0]])
                self.assertIsNone(grad_ss)

                with tf.Session() as sess:
                    sess.run(tf.global_variables_initializer())
                    grad_a, grad_s= sess.run([grad_a, grad_s])
                    self.assertTrue(np.sum(grad_a) != 0.0)
                    self.assertTrue(np.sum(grad_s) != 0.0)


class TestMarkovRecurrentModel(unittest.TestCase):

    def setUp(self):
        self.rddl1 = rddlgym.make('Navigation-v3', mode=rddlgym.AST)
        self.compiler1 = Compiler(self.rddl1, batch_mode=True)

        self.layers = [64, 32, 16]
        self.policy1 = DeepReactivePolicy(self.compiler1, self.layers, tf.nn.elu, input_layer_norm=False)

        self.batch_size1 = 100
        self.mrm1 = MarkovRecurrentModel(self.compiler1, self.policy1, self.batch_size1)

    def test_timesteps(self):
        horizon = 40
        simulators = [self.mrm1]
        batch_sizes = [self.batch_size1]

        for mrm, batch_size in zip(simulators, batch_sizes):

            with mrm.graph.as_default():
                timesteps = mrm._timesteps(horizon)

            self.assertIsInstance(timesteps, tf.Tensor)
            self.assertListEqual(timesteps.shape.as_list(), [batch_size, horizon, 1])

            with tf.Session(graph=mrm.graph) as sess:
                timesteps = sess.run(timesteps)

                for t in timesteps:
                    self.assertListEqual(list(t), list(np.arange(horizon-1, -1, -1)))

    def test_reparam_flags(self):
        horizon = 20
        simulators = [self.mrm1]
        batch_sizes = [self.batch_size1]

        for mrm, batch_size in zip(simulators, batch_sizes):

            with mrm.graph.as_default():

                reparam_flags1 = mrm._reparam_flags(horizon, ReparameterizationType.FULLY_REPARAMETERIZED)
                self.assertIsInstance(reparam_flags1, tf.Tensor)
                self.assertListEqual(reparam_flags1.shape.as_list(), [batch_size, horizon, 1])

                reparam_flags2 = mrm._reparam_flags(horizon, ReparameterizationType.NOT_REPARAMETERIZED)
                self.assertIsInstance(reparam_flags2, tf.Tensor)
                self.assertListEqual(reparam_flags2.shape.as_list(), [batch_size, horizon, 1])

                n_step = 5
                reparam_flags3 = mrm._reparam_flags(horizon, (ReparameterizationType.PARTIALLY_REPARAMETERIZED, n_step))
                self.assertIsInstance(reparam_flags3, tf.Tensor)
                self.assertListEqual(reparam_flags3.shape.as_list(), [batch_size, horizon, 1])

                not_reparam_constant = MarkovRecurrentModel.NOT_REPARAMETERIZED_FLAG
                fully_reparam_constant = MarkovRecurrentModel.FULLY_REPARAMETERIZED_FLAG
                with tf.Session(graph=mrm.graph) as sess:
                    f1, f2, f3 = sess.run([reparam_flags1, reparam_flags2, reparam_flags3])
                    self.assertTrue(np.all(f1 == fully_reparam_constant))
                    self.assertTrue(np.all(f2 == not_reparam_constant))

                    self.assertTrue(all(flag == not_reparam_constant for t, flag in enumerate(f3[0, :, 0]) if (t+1) % n_step == 0))
                    self.assertTrue(all(flag == fully_reparam_constant for t, flag in enumerate(f3[0, :, 0]) if (t+1) % n_step != 0))
                    for batch_flags in f3:
                        self.assertTrue(np.all(batch_flags == f3[0]))

    def test_inputs(self):
        horizon = 40
        simulators = [self.mrm1]
        batch_sizes = [self.batch_size1]

        for mrm, batch_size in zip(simulators, batch_sizes):

            with mrm.graph.as_default():

                timesteps = mrm._timesteps(horizon)

                flags1 = mrm._reparam_flags(horizon, ReparameterizationType.FULLY_REPARAMETERIZED)
                inputs1 = mrm._inputs(timesteps, flags1)
                self.assertIsInstance(inputs1, tf.Tensor)
                self.assertListEqual(inputs1.shape.as_list(), [batch_size, horizon, 2])

                flags2 = mrm._reparam_flags(horizon, ReparameterizationType.NOT_REPARAMETERIZED)
                inputs2 = mrm._inputs(timesteps, flags2)
                self.assertIsInstance(inputs2, tf.Tensor)
                self.assertListEqual(inputs2.shape.as_list(), [batch_size, horizon, 2])

                with tf.Session(graph=mrm.graph) as sess:
                    inputs1, inputs2 = sess.run([inputs1, inputs2])

                    for t in inputs1:
                        self.assertListEqual(list(t[:,0]), list(np.arange(horizon-1, -1, -1)))

                    for t in inputs2:
                        self.assertListEqual(list(t[:,0]), list(np.arange(horizon-1, -1, -1)))

                    self.assertTrue(np.all(inputs1[:,:,1] == np.zeros((batch_size, horizon))))
                    self.assertTrue(np.all(inputs2[:,:,1] == np.ones((batch_size, horizon))))

    def test_trajectory_fully_reparameterized(self):
        self._test_trajectory(ReparameterizationType.FULLY_REPARAMETERIZED)

    def test_trajectory_not_reparameterized(self):
        self._test_trajectory(ReparameterizationType.NOT_REPARAMETERIZED)

    def _test_trajectory(self, reparam_type):
        horizon = 40
        compilers = [self.compiler1]
        simulators = [self.mrm1]
        batch_sizes = [self.batch_size1]

        for compiler, mrm, batch_size in zip(compilers, simulators, batch_sizes):

            with mrm.graph.as_default():
                initial_state = mrm._cell.initial_state()

                timesteps = mrm._timesteps(horizon)
                flags = mrm._reparam_flags(horizon, reparam_type)
                inputs = mrm._inputs(timesteps, flags)

                trajectory = mrm._trajectory(initial_state, inputs)

            self.assertIsInstance(trajectory, tuple)
            self.assertEqual(len(trajectory), 5)

            # sizes
            state_size, action_size, interm_size, reward_size, log_prob_size = mrm.output_size

            # states
            self.assertIsInstance(trajectory.states, tuple)
            self.assertEqual(len(trajectory[1]), len(state_size))
            for s, sz in zip(trajectory[1], state_size):
                self.assertIsInstance(s, tf.Tensor)
                self.assertListEqual(s.shape.as_list(), [batch_size, horizon] + list(sz), '{}'.format(s))

            # interms
            self.assertIsInstance(trajectory.interms, tuple)
            self.assertEqual(len(trajectory.interms), len(interm_size))
            for s, sz in zip(trajectory.interms, interm_size):
                self.assertIsInstance(s, tf.Tensor)
                self.assertListEqual(s.shape.as_list(), [batch_size, horizon] + list(sz), '{}'.format(s))

            # actions
            self.assertIsInstance(trajectory.actions, tuple)
            self.assertEqual(len(trajectory.actions), len(action_size))
            for a, sz in zip(trajectory.actions, action_size):
                self.assertIsInstance(a, tf.Tensor)
                self.assertListEqual(a.shape.as_list(), [batch_size, horizon] + list(sz))

            # rewards
            self.assertIsInstance(trajectory.rewards, tf.Tensor)
            self.assertListEqual(trajectory.rewards.shape.as_list(), [batch_size, horizon, reward_size])

            # log_probs
            self.assertIsInstance(trajectory.log_probs, tf.Tensor)
            self.assertListEqual(trajectory.log_probs.shape.as_list(), [batch_size, horizon, log_prob_size])

    def test_reward_to_go(self):
        horizon = 40
        compilers = [self.compiler1]
        simulators = [self.mrm1]
        batch_sizes = [self.batch_size1]

        for compiler, mrm, batch_size in zip(compilers, simulators, batch_sizes):

            with mrm.graph.as_default():
                initial_state = mrm._cell.initial_state()

                timesteps = mrm._timesteps(horizon)
                flags = mrm._reparam_flags(horizon, ReparameterizationType.NOT_REPARAMETERIZED)
                inputs = mrm._inputs(timesteps, flags)

                trajectory = mrm._trajectory(initial_state, inputs)
                q = mrm._reward_to_go(trajectory.rewards)

            self.assertIsInstance(q, tf.Tensor)
            self.assertEqual(q.dtype, tf.float32)
            self.assertEqual(q.shape, trajectory.rewards.shape)

    def test_total_reward(self):
        horizon = 40
        compilers = [self.compiler1]
        simulators = [self.mrm1]
        batch_sizes = [self.batch_size1]

        for compiler, mrm, batch_size in zip(compilers, simulators, batch_sizes):
            mrm.build(horizon, lambda x: -x, ReparameterizationType.FULLY_REPARAMETERIZED)
            total_reward = mrm.total_reward
            self.assertIsInstance(total_reward, tf.Tensor)
            self.assertEqual(total_reward.dtype, tf.float32)
            self.assertListEqual(total_reward.shape.as_list(), [batch_size])

    def test_surrogate_loss_fully_reparameterized(self):
        surrogate_batch_cost, costs, q, log_probs = self._test_surrogate_loss(ReparameterizationType.FULLY_REPARAMETERIZED)
        self.assertTrue(np.all(surrogate_batch_cost == costs))

    def test_surrogate_loss_not_reparameterized(self):
        surrogate_batch_cost, costs, q, log_probs = self._test_surrogate_loss(ReparameterizationType.NOT_REPARAMETERIZED)
        self.assertTrue(np.all(surrogate_batch_cost == costs + log_probs * q))

    def _test_surrogate_loss(self, reparameterization_type):
        horizon = 40
        compilers = [self.compiler1]
        simulators = [self.mrm1]
        batch_sizes = [self.batch_size1]

        for compiler, mrm, batch_size in zip(compilers, simulators, batch_sizes):
            mrm.build(horizon, lambda x: -x, reparameterization_type)
            costs = mrm.costs
            q = mrm.q
            log_probs = mrm.trajectory.log_probs
            surrogate_batch_cost = mrm.surrogate_batch_cost

            with tf.Session(graph=mrm.graph) as sess:
                sess.run(tf.global_variables_initializer())
                return sess.run([surrogate_batch_cost, costs, q, log_probs])



# This file is part of tf-mdp.

# tf-mdp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-mdp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-mdp. If not, see <http://www.gnu.org/licenses/>.

import rddlgym

from tfmdp.train.policy import DeepReactivePolicy
from tfmdp.train.optimizer import PolicyOptimizer
from tfmdp.train.mrm import MarkovRecurrentModel, ReparameterizationType
from tfmdp.test.evaluator import PolicyEvaluator

import numpy as np
import tensorflow as tf

import unittest


class TestPolicyEvaluator(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # hyper-parameters
        cls.learning_rate = 0.001
        cls.epochs = 10
        cls.batch_size = 32
        cls.horizon = 20

        # model
        cls.compiler = rddlgym.make('Reservoir-8', mode=rddlgym.SCG)
        cls.compiler.batch_mode_on()

        # fluents
        cls.initial_state = cls.compiler.compile_initial_state(cls.batch_size)
        cls.state_fluents = cls.compiler.state_fluent_ordering
        cls.action_fluents = cls.compiler.action_fluent_ordering

        # policy
        cls.layers = [64, 32, 16]

        # optimizer
        cls.policy = DeepReactivePolicy(cls.compiler, cls.layers, tf.nn.elu, input_layer_norm=True)
        cls.model = MarkovRecurrentModel(cls.compiler, cls.policy, cls.batch_size)
        cls.model.build(cls.horizon, lambda x: -x, ReparameterizationType.FULLY_REPARAMETERIZED)
        cls.optimizer = PolicyOptimizer(cls.model)
        cls.optimizer.build(cls.learning_rate, cls.batch_size, cls.horizon, tf.train.RMSPropOptimizer)
        cls.optimizer.run(cls.epochs, show_progress=False)
        cls.save_path = cls.policy._checkpoint

        # evaluator
        cls.compiler = rddlgym.make('Reservoir-8', mode=rddlgym.SCG)
        cls.compiler.batch_mode_on()
        cls.policy = DeepReactivePolicy(cls.compiler, cls.layers, tf.nn.elu, input_layer_norm=True)
        cls.evaluator = PolicyEvaluator(cls.compiler, cls.policy)

    def test_policy_variables(self):

        with self.compiler.graph.as_default():

            input_layer_variables = tf.trainable_variables('trajectory/policy/input/')
            self.assertEqual(len(input_layer_variables), 2 * len(self.state_fluents))

            hidden_layer_variables = tf.trainable_variables('trajectory/policy/hidden')
            self.assertEqual(len(hidden_layer_variables), 2 * len(self.layers))

            output_layer_variables = tf.trainable_variables('trajectory/policy/output/')
            self.assertEqual(len(output_layer_variables), 2 * len(self.action_fluents))

    def test_evaluation_run(self):
        horizon = 45
        batch_size = 1024
        trajectories, _, _ = self.evaluator.run(horizon, batch_size, self.save_path)
        self.assertIsInstance(trajectories, tuple)
        self.assertEqual(len(trajectories), 5)

        initial_state, states, actions, interms, rewards = trajectories

        for fluent in initial_state:
            self.assertEqual(fluent.shape[0], batch_size)

        for fluent in states:
            self.assertTupleEqual(fluent.shape[:2], (batch_size, horizon))

        for fluent in actions:
            self.assertTupleEqual(fluent.shape[:2], (batch_size, horizon))

        for fluent in interms:
            self.assertTupleEqual(fluent.shape[:2], (batch_size, horizon))

        self.assertTupleEqual(rewards.shape[:2], (batch_size, horizon))



# This file is part of tf-mdp.

# tf-mdp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-mdp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-mdp. If not, see <http://www.gnu.org/licenses/>.


from tfmdp.train.policy import DeepReactivePolicy
from tfmdp.train.mrm import MarkovRecurrentModel, ReparameterizationType
from tfmdp.train.simulator import PolicySimulationModel
from tfmdp.train.optimizer import PolicyOptimizer

import tensorflow as tf


class PolicyOptimizationPlanner(object):

    _loss_fn ={
        'linear': lambda c: tf.abs(0 - c),
        'mse': lambda c: tf.square(0 - c),
        # 'huber': tf.losses.huber_loss
    }

    _non_linearities = {
        'none': None,
        'sigmoid': tf.sigmoid,
        'tanh': tf.tanh,
        'relu': tf.nn.relu,
        'relu6': tf.nn.relu6,
        'crelu': tf.nn.crelu,
        'elu': tf.nn.elu,
        'selu': tf.nn.selu,
        'softplus': tf.nn.softplus,
        'softsign': tf.nn.softsign
    }

    _optimizers = {
        'Adadelta': tf.train.AdadeltaOptimizer,
        'Adagrad': tf.train.AdagradOptimizer,
        'Adam': tf.train.AdamOptimizer,
        'GradientDescent': tf.train.GradientDescentOptimizer,
        'ProximalGradientDescent': tf.train.ProximalGradientDescentOptimizer,
        'ProximalAdagrad': tf.train.ProximalAdagradOptimizer,
        'RMSProp': tf.train.RMSPropOptimizer
    }

    def __init__(self,
            compiler,
            layers, activation, input_layer_norm, hidden_layer_norm,
            logdir=None):
        self._compiler = compiler
        self._logdir = logdir

        self._policy = DeepReactivePolicy(
            self._compiler,
            layers, self._non_linearities[activation],
            input_layer_norm, hidden_layer_norm)

    def build(self,
            learning_rate, batch_size, horizon,
            optimizer='RMSProp',
            loss='linear',
            kernel_l1_regularizer=None, kernel_l2_regularizer=None,
            bias_l1_regularizer=None, bias_l2_regularizer=None,
            reparameterization_type=None):

        self._model = MarkovRecurrentModel(self._compiler, self._policy, batch_size)
        if reparameterization_type is None:
            reparameterization_type = ReparameterizationType.FULLY_REPARAMETERIZED
        self._model.build(horizon, self._loss_fn[loss], reparameterization_type)

        self._optimizer = PolicyOptimizer(self._model, self._logdir, debug=False)
        self._optimizer.build(
            learning_rate, batch_size, horizon,
            self._optimizers[optimizer],
            self._get_regularizer(kernel_l1_regularizer, kernel_l2_regularizer),
            self._get_regularizer(bias_l1_regularizer, bias_l2_regularizer))

    def run(self, epochs, show_progress=True):
        losses, rewards = self._optimizer.run(epochs, show_progress=show_progress)
        logdir = self._optimizer._train_writer.get_logdir()
        return rewards, self._policy, logdir

    def _get_regularizer(self, l1_regularizer, l2_regularizer):
        regularizer = []
        if l1_regularizer != 0.0:
            regularizer.append(tf.contrib.layers.l1_regularizer(l1_regularizer))
        if l2_regularizer != 0.0:
            regularizer.append(tf.contrib.layers.l2_regularizer(l2_regularizer))
        if regularizer:
            return tf.contrib.layers.sum_regularizer(regularizer)



# This file is part of tf-mdp.

# tf-mdp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-mdp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-mdp. If not, see <http://www.gnu.org/licenses/>.


from tfmdp.train.policy import DeepReactivePolicy

from rddl2tf.compiler import Compiler
from tfrddlsim.simulation.policy_simulator import PolicySimulator

import time
import numpy as np
import tensorflow as tf

from typing import Sequence, Optional, Tuple
NonFluentsArray = Sequence[np.array]
StateArray = Sequence[np.array]
StatesArray = Sequence[np.array]
ActionsArray = Sequence[np.array]
IntermsArray = Sequence[np.array]
SimulationOutput = Tuple[NonFluentsArray, StateArray, StatesArray, ActionsArray, IntermsArray, np.array]


class PolicyEvaluator(object):
    '''PolicyEvaluator is a wraper around tfrddlsim.simulation.PolicySimulator
    for evaluating deep reactive policies.

    Args:
        compiler (:obj:`tfrddlsim.rddl2tf.compiler.Compiler`): A RDDL2TensorFlow compiler.
        policy (:obj:`tfmdp.train.policy.DeepReactivePolicy`): A deep reactive policy.
    '''

    def __init__(self, compiler: Compiler, policy: DeepReactivePolicy) -> None:
        self._compiler = compiler
        self._policy = policy

    @property
    def graph(self) -> tf.Graph:
        '''Returns the compiler's graph.'''
        return self._compiler.graph

    def run(self, horizon: int, batch_size: int, save_path: Optional[str] = None) -> SimulationOutput:
        '''Runs the trajectory simulation ops for the deep reactive policy.

        Returns:
            Tuple[StateArray, StatesArray, ActionsArray, IntermsArray, np.array]: Simulation output tuple.
        '''
        start = time.time()
        self._simulator = PolicySimulator(self._compiler, self._policy, batch_size)
        trajectories = self._simulator.trajectory(horizon)
        end = time.time()
        building_time = end - start

        start = time.time()
        with tf.Session(graph=self.graph) as sess:
            self._policy.restore(sess, save_path)
            trajectories_ = sess.run(trajectories)
        end = time.time()
        inference_time = end - start

        return trajectories_, building_time, inference_time



# This file is part of tf-mdp.

# tf-mdp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-mdp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-mdp. If not, see <http://www.gnu.org/licenses/>.

import rddlgym

from tfmdp.train.policy import DeepReactivePolicy

import numpy as np
import tensorflow as tf

import unittest


class TestDeepReactivePolicy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # hyper-parameters
        cls.batch_size = 64
        cls.horizon = 15

        # model
        cls.compiler = rddlgym.make('Reservoir-8', mode=rddlgym.SCG)
        cls.compiler.batch_mode_on()

        # initial state
        cls.initial_state = cls.compiler.compile_initial_state(cls.batch_size)

        # Deep Reactive Policy
        cls.layers = [128, 64]
        cls.policy = DeepReactivePolicy(cls.compiler, cls.layers, tf.nn.elu, input_layer_norm=True)
        cls.action = cls.policy(cls.initial_state, cls.horizon)

    def test_policy_name(self):
        actual = self.policy.name
        expected = 'drp-fc-layers={}'.format('+'.join(map(str, self.layers)))
        self.assertEqual(actual, expected)

    def test_state_inputs(self):
        self.assertIsInstance(self.policy.state_inputs, tuple)
        self.assertEqual(len(self.policy.state_inputs), len(self.initial_state))
        for layer, fluent in zip(self.policy.state_inputs, self.initial_state):
            self.assertIsInstance(layer, tf.Tensor)
            self.assertEqual(layer.dtype, fluent.dtype)
            self.assertEqual(len(layer.shape), 2)
            self.assertEqual(layer.shape.as_list()[0], self.batch_size)
            self.assertEqual(np.prod(layer.shape.as_list()), np.prod(fluent.shape.as_list()))

    def test_input_layer(self):
        self.assertIsInstance(self.policy.input_layer, tf.Tensor)

        sizes = list(np.prod(fluent.shape.as_list()) / self.batch_size for fluent in self.initial_state)
        total_size = sum(sizes)
        shape = [self.batch_size, total_size]
        self.assertListEqual(self.policy.input_layer.shape.as_list(), shape)

        with self.compiler.graph.as_default():
            state_fluents = self.compiler.state_fluent_ordering
            state_size = self.compiler.state_size
            for name, shape in zip(state_fluents, state_size):
                name = name.replace('/', '-')

                # layer norm
                beta = 'policy/input/{}/LayerNorm/beta'.format(name)
                beta = tf.trainable_variables(beta)
                self.assertEqual(len(beta), 1)
                self.assertListEqual(beta[0].shape.as_list(), list(shape))

                gamma = 'policy/input/{}/LayerNorm/gamma'.format(name)
                gamma = tf.trainable_variables(gamma)
                self.assertEqual(len(gamma), 1)
                self.assertListEqual(gamma[0].shape.as_list(), list(shape))

    def test_hidden_layers(self):
        self.assertIsInstance(self.policy.hidden, tuple)
        self.assertEqual(len(self.policy.hidden), len(self.layers) + 1)

        with self.compiler.graph.as_default():
            for l, units in enumerate(self.layers):
                vars = tf.trainable_variables('policy/hidden{}'.format(l+1))
                self.assertEqual(len(vars), 2)

                kernel = 'policy/hidden{}/dense/kernel'.format(l+1)
                vars = tf.trainable_variables(kernel)
                self.assertEqual(len(vars), 1)
                self.assertEqual(vars[0].shape[1], units)

                bias = 'policy/hidden{}/dense/bias'.format(l+1)
                vars = tf.trainable_variables(bias)
                self.assertEqual(len(vars), 1)
                self.assertEqual(vars[0].shape[0], units)

    def test_output_layer(self):
        action_fluents = self.compiler.action_fluent_ordering
        action_size = self.compiler.action_size
        self.assertIsInstance(self.policy.output_layer, tuple)
        self.assertEqual(len(self.policy.output_layer), len(action_fluents))

        with self.compiler.graph.as_default():

            for name, shape, layer in zip(action_fluents, action_size, self.policy.output_layer):
                self.assertIsInstance(layer, tf.Tensor)
                self.assertEqual(layer.dtype, tf.float32)
                self.assertEqual(len(layer.shape), 2)
                self.assertEqual(layer.shape[0], self.batch_size)
                self.assertEqual(layer.shape[1], np.prod(shape))

                name = name.replace('/', '-')
                vars = tf.trainable_variables('policy/output/{}'.format(name))
                self.assertEqual(len(vars), 2)

                kernel = 'policy/output/{}/dense/kernel'.format(name)
                vars = tf.trainable_variables(kernel)
                self.assertEqual(len(vars), 1)

                bias = 'policy/output/{}/dense/bias'.format(name)
                vars = tf.trainable_variables(bias)
                self.assertEqual(len(vars), 1)

    def test_action_outputs(self):
        action_fluents = self.compiler.action_fluent_ordering
        action_size = self.compiler.action_size
        action_dtype = self.compiler.action_dtype
        self.assertIsInstance(self.policy.action_outputs, tuple)
        self.assertEqual(len(self.policy.action_outputs), len(action_fluents))
        for name, shape, dtype, tensor in zip(action_fluents, action_size, action_dtype, self.policy.action_outputs):
            self.assertIsInstance(tensor, tf.Tensor)
            self.assertEqual(tensor.dtype, dtype)
            self.assertListEqual(tensor.shape.as_list(), [self.batch_size] + list(shape))



# This file is part of tf-mdp.

# tf-mdp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-mdp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-mdp. If not, see <http://www.gnu.org/licenses/>.

import rddlgym

from tfmdp.train.policy import DeepReactivePolicy
from tfmdp.train.mrm import MarkovRecurrentModel, ReparameterizationType
from tfmdp.train.optimizer import PolicyOptimizer

import numpy as np
import tensorflow as tf

import unittest


class TestPolicyOptimizer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # hyper-parameters
        cls.learning_rate = 0.0001
        cls.epochs = 32
        cls.batch_size = 64
        cls.horizon = 20

        # model
        cls.compiler = rddlgym.make('Reservoir-8', mode=rddlgym.SCG)
        cls.compiler.batch_mode_on()

        # fluents
        cls.initial_state = cls.compiler.compile_initial_state(cls.batch_size)
        cls.state_fluents = cls.compiler.state_fluent_ordering
        cls.state_size = cls.compiler.state_size
        cls.action_fluents = cls.compiler.action_fluent_ordering
        cls.action_size = cls.compiler.action_size

        # policy
        cls.layers = [64, 32, 16]
        cls.policy = DeepReactivePolicy(cls.compiler, cls.layers, tf.nn.elu, input_layer_norm=True)

        # model
        cls.model = MarkovRecurrentModel(cls.compiler, cls.policy, cls.batch_size)
        cls.model.build(cls.horizon, lambda x: -x, ReparameterizationType.FULLY_REPARAMETERIZED)

        # optimizer
        cls.optimizer = PolicyOptimizer(cls.model)
        cls.optimizer.build(cls.learning_rate, cls.batch_size, cls.horizon, tf.train.RMSPropOptimizer)

    def test_policy_variables(self):

        with self.compiler.graph.as_default():

            input_layer_variables = tf.trainable_variables('trajectory/policy/input/')
            self.assertEqual(len(input_layer_variables), 2 * len(self.state_fluents))

            hidden_layer_variables = tf.trainable_variables('trajectory/policy/hidden')
            self.assertEqual(len(hidden_layer_variables), 2 * len(self.layers))

            output_layer_variables = tf.trainable_variables('trajectory/policy/output/')
            self.assertEqual(len(output_layer_variables), 2 * len(self.action_fluents))

    def test_optimization_objective(self):
        loss = self.optimizer.loss
        self.assertIsInstance(loss, tf.Tensor, 'loss function is a tensor')
        self.assertEqual(loss.dtype, tf.float32, 'loss function is a real tensor')
        self.assertListEqual(loss.shape.as_list(), [], 'loss function is a scalar')

    def test_loss_optimizer(self):
        optimizer = self.optimizer._optimizer
        self.assertIsInstance(optimizer, tf.train.RMSPropOptimizer)
        train_op = self.optimizer._train_op
        self.assertEqual(train_op.name, 'policy_optimizer/RMSProp')

    def test_optimization_run(self):
        losses, rewards = self.optimizer.run(self.epochs, show_progress=False)
        self.assertIsInstance(self.policy._checkpoint, str)
        self.assertIsInstance(losses, list)
        self.assertIsInstance(rewards, list)
        self.assertEqual(len(losses), len(rewards))
        self.assertTrue(all(rewards[i+1] > rewards[i] for i in range(len(rewards[1:]))))



# This file is part of tf-mdp.

# tf-mdp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-mdp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-mdp. If not, see <http://www.gnu.org/licenses/>.

import rddlgym

from rddl2tf.compiler import Compiler

from tfmdp.train.policy import DeepReactivePolicy
from tfmdp.train.valuefn import Value

import numpy as np
import tensorflow as tf

import unittest

class TestValueFn(unittest.TestCase):

    def setUp(self):
        self.rddl1 = rddlgym.make('Navigation-v3', mode=rddlgym.AST)
        self.compiler1 = Compiler(self.rddl1, batch_mode=True)

        self.layers = [64]
        self.policy1 = DeepReactivePolicy(self.compiler1, self.layers, tf.nn.elu, input_layer_norm=False)

        self.valuefn1 = Value(self.compiler1, self.policy1)

    def test_build_trajectory_graph(self):
        horizon = 40
        batch_size = 128
        self.valuefn1.build(horizon, batch_size)

        states = self.valuefn1._states
        state_size = self.compiler1.state_size
        self.assertIsInstance(states, tuple)
        self.assertEqual(len(states), len(state_size))
        for fluent, size in zip(states, state_size):
            self.assertIsInstance(fluent, tf.Tensor)
            self.assertListEqual(fluent.shape.as_list(), [batch_size, horizon] + list(size))

        timesteps = self.valuefn1._timesteps
        self.assertIsInstance(timesteps, tf.Tensor)
        self.assertListEqual(timesteps.shape.as_list(), [batch_size, horizon])

    def test_build_value_estimates_graph(self):
        horizon = 40
        batch_size = 128
        self.valuefn1.build(horizon, batch_size)

        estimates = self.valuefn1._estimates
        self.assertIsInstance(estimates, tf.Tensor)
        self.assertListEqual(estimates.shape.as_list(), [batch_size, horizon])

    def test_regression_graph(self):
        horizon = 40
        batch_size = 128
        self.valuefn1.build(horizon, batch_size)

        features = self.valuefn1._dataset_features
        self.assertIsInstance(features, tuple)
        self.assertEqual(len(features), 2)

        timesteps, states = features
        self.assertIsInstance(timesteps, tf.Tensor)
        self.assertListEqual(timesteps.shape.as_list(), [batch_size * horizon])

        state_size = self.compiler1.state_size
        self.assertIsInstance(states, tuple)
        self.assertEqual(len(states), len(state_size))
        for fluent, size in zip(states, state_size):
            self.assertIsInstance(fluent, tf.Tensor)
            self.assertListEqual(fluent.shape.as_list(), [batch_size * horizon] + list(size))

        targets = self.valuefn1._dataset_targets
        self.assertIsInstance(targets, tf.Tensor)
        self.assertListEqual(targets.shape.as_list(), [batch_size * horizon])

    def test_prediction_graph(self):
        horizon = 40
        batch_size = 128
        self.valuefn1.build(horizon, batch_size)

        predictions = self.valuefn1._predictions
        self.assertIsInstance(predictions, tf.Tensor)
        self.assertListEqual(predictions.shape.as_list(), [None])

    def test_loss_graph(self):
        horizon = 40
        batch_size = 128
        self.valuefn1.build(horizon, batch_size)

        loss = self.valuefn1._loss

    def test_optimization_graph(self):
        horizon = 20
        batch_size = 4
        self.valuefn1.build(horizon, batch_size)

        optimizer = self.valuefn1._optimizer
        grad_and_vars = self.valuefn1._grad_and_vars
        train_op = self.valuefn1._train_op

    def test_training_batch(self):
        horizon = 5
        batch_size = 4
        self.valuefn1.build(horizon, batch_size)

        with tf.Session(graph=self.compiler1.graph) as sess:
            sess.run(tf.global_variables_initializer())

            training_batch = 2
            dataset = self.valuefn1._regression_dataset(sess)
            batchs = self.valuefn1._training_batch_generator(training_batch, dataset)

            state_size = self.compiler1.state_size

            n = 0
            for state, timestep, target in batchs:
                self.assertIsInstance(state, tuple)
                self.assertEqual(len(state), len(state_size))

                for fluent, size in zip(state, state_size):
                    self.assertIsInstance(fluent, np.ndarray)
                    self.assertListEqual(list(fluent.shape), [training_batch] + list(size))

                self.assertIsInstance(timestep, np.ndarray)
                self.assertListEqual(list(timestep.shape), [training_batch])

                self.assertIsInstance(target, np.ndarray)
                self.assertListEqual(list(target.shape), [training_batch])

                n += 1

            self.assertEqual(n, int(batch_size * horizon / training_batch))

    def test_value_fn_fitting(self):
        horizon = 40
        batch_size = 128
        self.valuefn1.build(horizon, batch_size)

        with tf.Session(graph=self.compiler1.graph) as sess:
            sess.run(tf.global_variables_initializer())

            epochs = 20
            batch_size = 64
            loss = self.valuefn1.fit(sess, batch_size, epochs, show_progress=True)
            self.assertIsInstance(loss, list)

